@@ -74,6 +74,8 @@ public:
   nsGenericHTMLElement(nsINodeInfo *aNodeInfo)
     : nsGenericHTMLElementBase(aNodeInfo)
   {
+    NS_ASSERTION(aNodeInfo->NamespaceID() == kNameSpaceID_XHTML,
+                 "Unexpected namespace");
   }
 
   /** Typesafe, non-refcounting cast from nsIContent.  Cheaper than QI. **/
@@ -103,7 +105,6 @@ public:
 
   // Implementation for nsIDOMNode
   NS_METHOD GetNodeName(nsAString& aNodeName);
-  NS_METHOD GetLocalName(nsAString& aLocalName);
 
   // Implementation for nsIDOMElement
   NS_METHOD SetAttribute(const nsAString& aName,
@@ -111,9 +112,6 @@ public:
   NS_METHOD GetTagName(nsAString& aTagName);
   NS_METHOD GetElementsByTagName(const nsAString& aTagname,
                                  nsIDOMNodeList** aReturn);
-  NS_METHOD GetElementsByTagNameNS(const nsAString& aNamespaceURI,
-                                   const nsAString& aLocalName,
-                                   nsIDOMNodeList** aReturn);
 
   // nsIDOMHTMLElement methods. Note that these are non-virtual
   // methods, implementations are expected to forward calls to these
@@ -176,7 +174,6 @@ public:
   virtual nsresult UnsetAttr(PRInt32 aNameSpaceID, nsIAtom* aName,
                              PRBool aNotify);
   virtual PRBool IsNodeOfType(PRUint32 aFlags) const;
-  virtual void RemoveFocus(nsPresContext *aPresContext);
   virtual PRBool IsFocusable(PRInt32 *aTabIndex = nsnull)
   {
     PRBool isFocusable = PR_FALSE;
@@ -201,15 +198,15 @@ public:
   PRBool IsHTMLLink(nsIURI** aURI) const;
 
   // Used by A, AREA, LINK, and STYLE.
-  // Callers must hold a reference to nsHTMLUtils's global reference count.
-  nsresult GetHrefURIForAnchors(nsIURI** aURI) const;
+  already_AddRefed<nsIURI> GetHrefURIForAnchors() const;
+
+  // As above, but makes sure to return a URI object that we can mutate with
+  // impunity without changing our current URI.  That is, if the URI is cached
+  // it clones it and returns the clone.
+  void GetHrefURIToMutate(nsIURI** aURI);
 
   // HTML element methods
   void Compact() { mAttrsAndChildren.Compact(); }
-  const nsAttrValue* GetParsedAttr(nsIAtom* aAttr) const
-  {
-    return mAttrsAndChildren.GetAttr(aAttr);
-  }
 
   virtual void UpdateEditableState();
 
@@ -520,63 +517,30 @@ public:
   static PRBool InNavQuirksMode(nsIDocument* aDoc);
 
   // Helper functions for <a> and <area>
-  static nsresult SetProtocolInHrefString(const nsAString &aHref,
-                                          const nsAString &aProtocol,
-                                          nsAString &aResult);
+  void SetHrefToURI(nsIURI* aURI);
+  nsresult SetProtocolInHrefURI(const nsAString &aProtocol);
+  nsresult SetHostInHrefURI(const nsAString &aHost);
+  nsresult SetHostnameInHrefURI(const nsAString &aHostname);
+  nsresult SetPathnameInHrefURI(const nsAString &aPathname);
+  nsresult SetSearchInHrefURI(const nsAString &aSearch);
+  nsresult SetPortInHrefURI(const nsAString &aPort);
+  nsresult SetHashInHrefURI(const nsAString &aHash);
 
-  static nsresult SetHostInHrefString(const nsAString &aHref,
-                                      const nsAString &aHost,
-                                      nsAString &aResult);
+  nsresult GetProtocolFromHrefURI(nsAString& aProtocol);
+  nsresult GetHostFromHrefURI(nsAString& aHost);
+  nsresult GetHostnameFromHrefURI(nsAString& aHostname);
+  nsresult GetPathnameFromHrefURI(nsAString& aPathname);
+  nsresult GetSearchFromHrefURI(nsAString& aSearch);
+  nsresult GetPortFromHrefURI(nsAString& aPort);
+  nsresult GetHashFromHrefURI(nsAString& aHash);
 
-  static nsresult SetHostnameInHrefString(const nsAString &aHref,
-                                          const nsAString &aHostname,
-                                          nsAString &aResult);
-
-  static nsresult SetPathnameInHrefString(const nsAString &aHref,
-                                          const nsAString &aHostname,
-                                          nsAString &aResult);
-
-  static nsresult SetSearchInHrefString(const nsAString &aHref,
-                                        const nsAString &aSearch,
-                                        nsAString &aResult);
-  
-  static nsresult SetHashInHrefString(const nsAString &aHref,
-                                      const nsAString &aHash,
-                                      nsAString &aResult);
-
-  static nsresult SetPortInHrefString(const nsAString &aHref,
-                                      const nsAString &aPort,
-                                      nsAString &aResult);
-
-  static nsresult GetProtocolFromHrefString(const nsAString &aHref,
-                                            nsAString& aProtocol,
-                                            nsIDocument *aDocument);
-
-  static nsresult GetHostFromHrefString(const nsAString &aHref,
-                                        nsAString& aHost);
-
-  static nsresult GetHostnameFromHrefString(const nsAString &aHref,
-                                            nsAString& aHostname);
-
-  static nsresult GetPathnameFromHrefString(const nsAString &aHref,
-                                            nsAString& aPathname);
-
-  static nsresult GetSearchFromHrefString(const nsAString &aHref,
-                                          nsAString& aSearch);
-
-  static nsresult GetPortFromHrefString(const nsAString &aHref,
-                                        nsAString& aPort);
-
-  static nsresult GetHashFromHrefString(const nsAString &aHref,
-                                        nsAString& aHash);
-protected:
   /**
-   * Focus or blur the element.  This is what you should call if you want to
-   * *cause* a focus or blur on your element.  SetFocus / SetBlur are the
-   * methods where you want to catch what occurs on your element.
-   * @param aDoFocus true to focus, false to blur
+   * Locate an nsIEditor rooted at this content node, if there is one.
    */
-  void SetElementFocus(PRBool aDoFocus);
+  NS_HIDDEN_(nsresult) GetEditor(nsIEditor** aEditor);
+  NS_HIDDEN_(nsresult) GetEditorInternal(nsIEditor** aEditor);
+
+protected:
   /**
    * Register or unregister an access key to this element based on the
    * accesskey attribute.
@@ -714,6 +678,18 @@ protected:
    */
   NS_HIDDEN_(nsresult) GetURIAttr(nsIAtom* aAttr, nsIAtom* aBaseAttr, nsAString& aResult);
 
+  /**
+   * Helper for GetURIAttr and GetHrefURIForAnchors which returns an
+   * nsIURI in the out param.
+   *
+   * @param aCloneIfCached if true, clone the URI before returning if
+   * it's cached.
+   *
+   * @return PR_TRUE if we had the attr, PR_FALSE otherwise.
+   */
+  NS_HIDDEN_(PRBool) GetURIAttr(nsIAtom* aAttr, nsIAtom* aBaseAttr,
+                                PRBool aCloneIfCached, nsIURI** aURI) const;
+
   /**
    * This method works like GetURIAttr, except that it supports multiple
    * URIs separated by whitespace (one or more U+0020 SPACE characters).
@@ -728,12 +704,6 @@ protected:
    */
   NS_HIDDEN_(nsresult) GetURIListAttr(nsIAtom* aAttr, nsAString& aResult);
 
-  /**
-   * Locate an nsIEditor rooted at this content node, if there is one.
-   */
-  NS_HIDDEN_(nsresult) GetEditor(nsIEditor** aEditor);
-  NS_HIDDEN_(nsresult) GetEditorInternal(nsIEditor** aEditor);
-
   /**
    * Locates the nsIEditor associated with this node.  In general this is
    * equivalent to GetEditorInternal(), but for designmode or contenteditable,
@@ -781,6 +751,9 @@ protected:
     static const nsIContent::AttrValuesArray values[] =
       { &nsGkAtoms::_false, &nsGkAtoms::_true, &nsGkAtoms::_empty, nsnull };
 
+    if (!HasFlag(NODE_MAY_HAVE_CONTENT_EDITABLE_ATTR))
+      return eInherit;
+
     PRInt32 value = FindAttrValueIn(kNameSpaceID_None,
                                     nsGkAtoms::contenteditable, values,
                                     eIgnoreCase);
@@ -852,6 +825,8 @@ public:
   virtual PRUint32 GetDesiredIMEState();
   virtual PRInt32 IntrinsicState() const;
 
+  virtual nsresult PreHandleEvent(nsEventChainPreVisitor& aVisitor);
+
 protected:
   virtual nsresult BeforeSetAttr(PRInt32 aNameSpaceID, nsIAtom* aName,
                                  const nsAString* aValue, PRBool aNotify);
@@ -866,11 +841,6 @@ protected:
 
   void UpdateEditableFormControlState();
 
-  void SetFocusAndScrollIntoView(nsPresContext* aPresContext);
-
-  // A sane SetFocus implementation for focusable form controls
-  void DoSetFocus(nsPresContext* aPresContext);
-
   // The focusability state of this form control.  eUnfocusable means that it
   // shouldn't be focused at all, eInactiveWindow means it's in an inactive
   // window, eActiveWindow means it's in an active window.
