@@ -45,8 +45,10 @@
 
 #include "nsUnicharUtils.h"
 
-#include "nsIPref.h"
+#include "nsIPrefService.h"
+#include "nsIPrefBranch2.h"
 #include "nsServiceManagerUtils.h"
+#include "nsTArray.h"
 
 #include "nsIWindowsRegKey.h"
 #include "nsILocalFile.h"
@@ -57,12 +59,20 @@
 #include FT_FREETYPE_H
 #include "gfxFT2Fonts.h"
 #include "cairo-ft.h"
+#include "nsAppDirectoryServiceDefs.h"
 #else
 #include "gfxWindowsFonts.h"
 #endif
 
+/*XXX to get CAIRO_HAS_DDRAW_SURFACE */
+#include "cairo.h"
+
 #ifdef WINCE
 #include <shlwapi.h>
+
+#ifdef CAIRO_HAS_DDRAW_SURFACE
+#include "gfxDDrawSurface.h"
+#endif
 #endif
 
 #include "gfxUserFontSet.h"
@@ -86,14 +96,24 @@ BuildKeyNameFromFontName(nsAString &aName)
     ToLowerCase(aName);
 }
 
-int
-gfxWindowsPlatform::PrefChangedCallback(const char *aPrefName, void *closure)
+class gfxWindowsPlatformPrefObserver : public nsIObserver {
+public:
+    NS_DECL_ISUPPORTS
+    NS_DECL_NSIOBSERVER
+};
+
+NS_IMPL_ISUPPORTS1(gfxWindowsPlatformPrefObserver, nsIObserver)
+
+NS_IMETHODIMP
+gfxWindowsPlatformPrefObserver::Observe(nsISupports     *aSubject,
+                                        const char      *aTopic,
+                                        const PRUnichar *aData)
 {
+    NS_ASSERTION(!strcmp(aTopic, NS_PREFBRANCH_PREFCHANGE_TOPIC_ID), "invalid topic");
     // XXX this could be made to only clear out the cache for the prefs that were changed
     // but it probably isn't that big a deal.
-    gfxWindowsPlatform *plat = static_cast<gfxWindowsPlatform *>(closure);
-    plat->mPrefFonts.Clear();
-    return 0;
+    gfxWindowsPlatform::GetPlatform()->ClearPrefFonts();
+    return NS_OK;
 }
 
 gfxWindowsPlatform::gfxWindowsPlatform()
@@ -104,19 +124,47 @@ gfxWindowsPlatform::gfxWindowsPlatform()
     mFontSubstitutes.Init(50);
     mPrefFonts.Init(10);
 
-    UpdateFontList();
-
-    nsCOMPtr<nsIPref> pref = do_GetService(NS_PREF_CONTRACTID);
-    pref->RegisterCallback("font.", PrefChangedCallback, this);
-    pref->RegisterCallback("font.name-list.", PrefChangedCallback, this);
-    pref->RegisterCallback("intl.accept_languages", PrefChangedCallback, this);
-    // don't bother unregistering.  We'll get shutdown after the pref service
-
 #ifdef MOZ_FT2_FONTS
     FT_Init_FreeType(&gPlatformFTLibrary);
 #else
     FontEntry::InitializeFontEmbeddingProcs();
 #endif
+
+    UpdateFontList();
+
+    nsCOMPtr<nsIPrefBranch2> pref = do_GetService(NS_PREFSERVICE_CONTRACTID);
+
+    if (pref) {
+        gfxWindowsPlatformPrefObserver *observer = new gfxWindowsPlatformPrefObserver();
+        if (observer) {
+            pref->AddObserver("font.", observer, PR_FALSE);
+            pref->AddObserver("font.name-list.", observer, PR_FALSE);
+            pref->AddObserver("intl.accept_languages", observer, PR_FALSE);
+            // don't bother unregistering.  We'll get shutdown after the pref service
+        }
+    }
+
+/* Pick the default render mode differently between
+ * desktop, Windows Mobile, and Windows CE.
+ */
+#if defined(WINCE_WINDOWS_MOBILE)
+    mRenderMode = RENDER_IMAGE_DDRAW16;
+#elif defined(WINCE)
+    mRenderMode = RENDER_DDRAW_GL;
+#else
+    mRenderMode = RENDER_GDI;
+#endif
+
+    PRInt32 rmode;
+    if (NS_SUCCEEDED(pref->GetIntPref("mozilla.widget.render-mode", &rmode))) {
+        if (rmode >= 0 || rmode < RENDER_MODE_MAX) {
+#ifndef CAIRO_HAS_DDRAW_SURFACE
+            if (rmode == RENDER_DDRAW || rmode == RENDER_DDRAW_GL)
+                rmode = RENDER_IMAGE_STRETCH24;
+#endif
+            mRenderMode = (RenderMode) rmode;
+        }
+    }
 }
 
 gfxWindowsPlatform::~gfxWindowsPlatform()
@@ -129,12 +177,23 @@ already_AddRefed<gfxASurface>
 gfxWindowsPlatform::CreateOffscreenSurface(const gfxIntSize& size,
                                            gfxASurface::gfxImageFormat imageFormat)
 {
-#ifndef WINCE
-    gfxASurface *surf = new gfxWindowsSurface(size, imageFormat);
-#else
-    gfxASurface *surf = new gfxImageSurface(size, imageFormat);
+    gfxASurface *surf = nsnull;
+
+#ifdef CAIRO_HAS_DDRAW_SURFACE
+    if (mRenderMode == RENDER_DDRAW || mRenderMode == RENDER_DDRAW_GL)
+        surf = new gfxDDrawSurface(NULL, size, imageFormat);
 #endif
+
+#ifdef CAIRO_HAS_WIN32_SURFACE
+    if (mRenderMode == RENDER_GDI)
+        surf = new gfxWindowsSurface(size, imageFormat);
+#endif
+
+    if (surf == nsnull)
+        surf = new gfxImageSurface(size, imageFormat);
+
     NS_IF_ADDREF(surf);
+
     return surf;
 }
 
@@ -168,11 +227,11 @@ gfxWindowsPlatform::FontEnumProc(const ENUMLOGFONTEXW *lpelfe,
 // general cmap reading routines moved to gfxFontUtils.cpp
 
 struct FontListData {
-    FontListData(const nsACString& aLangGroup, const nsACString& aGenericFamily, nsStringArray& aListOfFonts) :
+    FontListData(const nsACString& aLangGroup, const nsACString& aGenericFamily, nsTArray<nsString>& aListOfFonts) :
         mLangGroup(aLangGroup), mGenericFamily(aGenericFamily), mStringArray(aListOfFonts) {}
     const nsACString& mLangGroup;
     const nsACString& mGenericFamily;
-    nsStringArray& mStringArray;
+    nsTArray<nsString>& mStringArray;
 };
 
 PLDHashOperator
@@ -192,6 +251,7 @@ gfxWindowsPlatform::HashEnumFunc(nsStringHashKey::KeyType aKey,
     if (!aFontEntry)
         return PL_DHASH_NEXT;
 
+
 #ifndef MOZ_FT2_FONTS
     /* skip symbol fonts */
     if (aFontEntry->mSymbolFont)
@@ -200,7 +260,7 @@ gfxWindowsPlatform::HashEnumFunc(nsStringHashKey::KeyType aKey,
     if (aFontEntry->SupportsLangGroup(data->mLangGroup) &&
         aFontEntry->MatchesGenericFamily(data->mGenericFamily))
 #endif
-        data->mStringArray.AppendString(aFontFamily->mName);
+    data->mStringArray.AppendElement(aFontFamily->Name());
 
     return PL_DHASH_NEXT;
 }
@@ -208,7 +268,7 @@ gfxWindowsPlatform::HashEnumFunc(nsStringHashKey::KeyType aKey,
 nsresult
 gfxWindowsPlatform::GetFontList(const nsACString& aLangGroup,
                                 const nsACString& aGenericFamily,
-                                nsStringArray& aListOfFonts)
+                                nsTArray<nsString>& aListOfFonts)
 {
     FontListData data(aLangGroup, aGenericFamily, aListOfFonts);
 
@@ -249,7 +309,8 @@ void gfxWindowsPlatform::AppendFacesFromFontFile(const PRUnichar *aFileName) {
                     ff = new FontFamily(name);
                     mFonts.Put(name, ff);
                 }
-                ff->mFaces.AppendElement(fe);
+                ff->AddFontEntry(fe);
+                ff->SetHasStyles(PR_TRUE);
             }
         }
         FT_Done_Face(dummy);
@@ -259,24 +320,36 @@ void gfxWindowsPlatform::AppendFacesFromFontFile(const PRUnichar *aFileName) {
 void
 gfxWindowsPlatform::FindFonts()
 {
-    nsTArray<nsString> searchPaths(2);
-    nsTArray<nsString> fontPatterns(4);
-    fontPatterns.AppendElement(NS_LITERAL_STRING("\\*.TTF"));
+    nsTArray<nsString> searchPaths(3);
+    nsTArray<nsString> fontPatterns(3);
     fontPatterns.AppendElement(NS_LITERAL_STRING("\\*.ttf"));
-    fontPatterns.AppendElement(NS_LITERAL_STRING("\\*.OTF"));
+    fontPatterns.AppendElement(NS_LITERAL_STRING("\\*.ttc"));
     fontPatterns.AppendElement(NS_LITERAL_STRING("\\*.otf"));
     wchar_t pathBuf[256];
     SHGetSpecialFolderPathW(0, pathBuf, CSIDL_WINDOWS, 0);
     searchPaths.AppendElement(pathBuf);
     SHGetSpecialFolderPathW(0, pathBuf, CSIDL_FONTS, 0);
     searchPaths.AppendElement(pathBuf);
+    nsCOMPtr<nsIFile> resDir;
+    NS_GetSpecialDirectory(NS_APP_RES_DIR, getter_AddRefs(resDir));
+    if (resDir) {
+        resDir->Append(NS_LITERAL_STRING("fonts"));
+        nsAutoString resPath;
+        resDir->GetPath(resPath);
+        searchPaths.AppendElement(resPath);
+    }
     WIN32_FIND_DATAW results;
     for (PRUint32 i = 0;  i < searchPaths.Length(); i++) {
         const nsString& path(searchPaths[i]);
         for (PRUint32 j = 0; j < fontPatterns.Length(); j++) { 
             nsAutoString pattern(path);
             pattern.Append(fontPatterns[j]);
-            HANDLE handle = FindFirstFileW(pattern.get(), &results);
+            HANDLE handle = FindFirstFileExW(pattern.get(),
+                                             FindExInfoStandard,
+                                             &results,
+                                             FindExSearchNameMatch,
+                                             NULL,
+                                             0);
             PRBool moreFiles = handle != INVALID_HANDLE_VALUE;
             while (moreFiles) {
                 nsAutoString filePath(path);
@@ -285,6 +358,8 @@ gfxWindowsPlatform::FindFonts()
                 AppendFacesFromFontFile(static_cast<const PRUnichar*>(filePath.get()));
                 moreFiles = FindNextFile(handle, &results);
             }
+            if (handle != INVALID_HANDLE_VALUE)
+                FindClose(handle);
         }
     }
 }
@@ -359,7 +434,7 @@ gfxWindowsPlatform::UpdateFontList()
         if (!actualFontName.IsEmpty() && mFonts.Get(actualFontName, &ff))
             mFontSubstitutes.Put(substituteName, ff);
         else
-            mNonExistingFonts.AppendString(substituteName);
+            mNonExistingFonts.AppendElement(substituteName);
     }
 
     // initialize ranges of characters for which system-wide font search should be skipped
@@ -461,12 +536,12 @@ gfxWindowsPlatform::ResolveFontName(const nsAString& aFontName,
     if (mFonts.Get(keyName, &ff) ||
         mFontSubstitutes.Get(keyName, &ff) ||
         mFontAliases.Get(keyName, &ff)) {
-        aAborted = !(*aCallback)(ff->mName, aClosure);
+        aAborted = !(*aCallback)(ff->Name(), aClosure);
         // XXX If the font has font link, we should add the linked font.
         return NS_OK;
     }
 
-    if (mNonExistingFonts.IndexOf(keyName) >= 0) {
+    if (mNonExistingFonts.Contains(keyName)) {
         aAborted = PR_FALSE;
         return NS_OK;
     }
@@ -487,7 +562,7 @@ gfxWindowsPlatform::ResolveFontName(const nsAString& aFontName,
                                     (FONTENUMPROCW)gfxWindowsPlatform::FontResolveProc,
                                     (LPARAM)&data, 0);
     if (data.mFoundCount == 0)
-        mNonExistingFonts.AppendString(keyName);
+        mNonExistingFonts.AppendElement(keyName);
     ::ReleaseDC(nsnull, dc);
 
     return NS_OK;
@@ -528,16 +603,6 @@ gfxWindowsPlatform::FontResolveProc(const ENUMLOGFONTEXW *lpelfe,
     // XXX If the font has font link, we should add the linked font.
 }
 
-struct FontSearch {
-    FontSearch(PRUint32 aCh, gfxFont *aFont) :
-        ch(aCh), fontToMatch(aFont), matchRank(-1) {
-    }
-    PRUint32 ch;
-    nsRefPtr<gfxFont> fontToMatch;
-    PRInt32 matchRank;
-    nsRefPtr<FontEntry> bestMatch;
-};
-
 PLDHashOperator
 gfxWindowsPlatform::FindFontForCharProc(nsStringHashKey::KeyType aKey,
                                         nsRefPtr<FontFamily>& aFontFamily,
@@ -545,16 +610,22 @@ gfxWindowsPlatform::FindFontForCharProc(nsStringHashKey::KeyType aKey,
 {
     FontSearch *data = (FontSearch*)userArg;
 
-    const PRUint32 ch = data->ch;
+#ifdef MOZ_FT2_FONTS
+    aFontFamily->FindFontForChar(data);
+#else
+    const PRUint32 ch = data->mCh;
 
-    nsRefPtr<FontEntry> fe = aFontFamily->FindFontEntry(*data->fontToMatch->GetStyle());
-    PRInt32 rank = 0;
+    nsRefPtr<FontEntry> fe = aFontFamily->FindFontEntry(*data->mFontToMatch->GetStyle());
     NS_ASSERTION(fe, "couldn't find any font entry in family");
-
-#ifndef MOZ_FT2_FONTS
     if (!fe)
         return PL_DHASH_NEXT;
 
+    // initialize rank to 1 so that any match is better than the initial
+    // value of data->mMatchRank (zero); therefore the first font that
+    // passes the mCharacterMap.test() will become the mBestMatch until
+    // a better entry is found
+    PRInt32 rank = 1;
+
     // skip over non-unicode and bitmap fonts and fonts that don't have
     // the code point we're looking for
     if (fe->IsCrappyFont() || !fe->mCharacterMap.test(ch))
@@ -565,35 +636,35 @@ gfxWindowsPlatform::FindFontForCharProc(nsStringHashKey::KeyType aKey,
     if (fe->SupportsRange(gfxFontUtils::CharRangeBit(ch)))
         rank += 1;
 
-    if (fe->SupportsLangGroup(data->fontToMatch->GetStyle()->langGroup))
+    if (fe->SupportsLangGroup(data->mFontToMatch->GetStyle()->langGroup))
         rank += 2;
 
-    FontEntry* mfe = static_cast<FontEntry*>(data->fontToMatch->GetFontEntry());
+    FontEntry* mfe = static_cast<FontEntry*>(data->mFontToMatch->GetFontEntry());
 
     if (fe->mWindowsFamily == mfe->mWindowsFamily)
         rank += 3;
     if (fe->mWindowsPitch == mfe->mWindowsPitch)
         rank += 3;
-#endif
 
     /* italic */
-    const PRBool italic = (data->fontToMatch->GetStyle()->style != FONT_STYLE_NORMAL);
+    const PRBool italic = (data->mFontToMatch->GetStyle()->style != FONT_STYLE_NORMAL);
     if (fe->mItalic != italic)
         rank += 3;
 
     /* weight */
     PRInt8 baseWeight, weightDistance;
-    data->fontToMatch->GetStyle()->ComputeWeightAndOffset(&baseWeight, &weightDistance);
+    data->mFontToMatch->GetStyle()->ComputeWeightAndOffset(&baseWeight, &weightDistance);
     if (fe->mWeight == (baseWeight * 100) + (weightDistance * 100))
         rank += 2;
-    else if (fe->mWeight == data->fontToMatch->GetFontEntry()->mWeight)
+    else if (fe->mWeight == data->mFontToMatch->GetFontEntry()->mWeight)
         rank += 1;
 
-    if (rank > data->matchRank ||
-        (rank == data->matchRank && Compare(fe->Name(), data->bestMatch->Name()) > 0)) {
-        data->bestMatch = fe;
-        data->matchRank = rank;
+    if (rank > data->mMatchRank ||
+        (rank == data->mMatchRank && Compare(fe->Name(), data->mBestMatch->Name()) > 0)) {
+        data->mBestMatch = fe;
+        data->mMatchRank = rank;
     }
+#endif
 
     return PL_DHASH_NEXT;
 }
@@ -611,16 +682,17 @@ gfxWindowsPlatform::FindFontForChar(PRUint32 aCh, gfxFont *aFont)
     // find fonts that support the character
     mFonts.Enumerate(gfxWindowsPlatform::FindFontForCharProc, &data);
 
-    if (data.bestMatch) {
+    if (data.mBestMatch) {
 #ifdef MOZ_FT2_FONTS
         nsRefPtr<gfxFT2Font> font =
-            gfxFT2Font::GetOrMakeFont(data.bestMatch->mName, 
+            gfxFT2Font::GetOrMakeFont(static_cast<FontEntry*>(data.mBestMatch.get()), 
                                       aFont->GetStyle()); 
-            gfxFont* ret = font.forget().get();
-            return already_AddRefed<gfxFont>(ret);
+        gfxFont* ret = font.forget().get();
+        return already_AddRefed<gfxFont>(ret);
 #else
         nsRefPtr<gfxWindowsFont> font =
-            gfxWindowsFont::GetOrMakeFont(data.bestMatch, aFont->GetStyle());
+            gfxWindowsFont::GetOrMakeFont(static_cast<FontEntry*>(data.mBestMatch.get()),
+                                          aFont->GetStyle());
         if (font->IsValid()) {
             gfxFont* ret = font.forget().get();
             return already_AddRefed<gfxFont>(ret);
@@ -709,13 +781,13 @@ FindFullName(nsStringHashKey::KeyType aKey,
     // if so, iterate over faces in this family to see if there is a match
     if (family.Equals(fullNameFamily)) {
 #ifdef MOZ_FT2_FONTS
-        int len = aFontFamily->mFaces.Length();
+        int len = aFontFamily->GetFontList().Length();
         int index = 0;
         for (; index < len && 
-                 !aFontFamily->mFaces[index]->Name().Equals(data->mFullName); index++);
+                 !aFontFamily->GetFontList()[index]->Name().Equals(data->mFullName); index++);
         if (index < len) {
             data->mFound = PR_TRUE;
-            data->mFontEntry = aFontFamily->mFaces[index];
+            data->mFontEntry = aFontFamily->GetFontList()[index];
         }
 #else
         HDC hdc;
@@ -769,13 +841,18 @@ gfxWindowsPlatform::LookupLocalFont(const gfxProxyFontEntry *aProxyEntry,
 
 gfxFontEntry* 
 gfxWindowsPlatform::MakePlatformFont(const gfxProxyFontEntry *aProxyEntry,
-                                     nsISupports *aLoader,
                                      const PRUint8 *aFontData, PRUint32 aLength)
 {
 #ifdef MOZ_FT2_FONTS
-    return FontEntry::CreateFontEntry(*aProxyEntry, aLoader, aFontData, aLength);
+    // The FT2 font needs the font data to persist, so we do NOT free it here
+    // but instead pass ownership to the font entry.
+    // Deallocation will happen later, when the font face is destroyed.
+    return FontEntry::CreateFontEntry(*aProxyEntry, aFontData, aLength);
 #else
-    return FontEntry::LoadFont(*aProxyEntry, aLoader, aFontData, aLength);
+    // With GDI, we can free the downloaded data after activating the font
+    gfxFontEntry *fe = FontEntry::LoadFont(*aProxyEntry, aFontData, aLength);
+    NS_Free((void*)aFontData);
+    return fe;
 #endif    
 }
 
@@ -787,7 +864,8 @@ gfxWindowsPlatform::IsFontFormatSupported(nsIURI *aFontURI, PRUint32 aFormatFlag
                  "strange font format hint set");
 
     // accept supported formats
-    if (aFormatFlags & (gfxUserFontSet::FLAG_FORMAT_OPENTYPE | 
+    if (aFormatFlags & (gfxUserFontSet::FLAG_FORMAT_WOFF     |
+                        gfxUserFontSet::FLAG_FORMAT_OPENTYPE | 
                         gfxUserFontSet::FLAG_FORMAT_TRUETYPE)) {
         return PR_TRUE;
     }
@@ -903,3 +981,14 @@ gfxWindowsPlatform::GetFTLibrary()
     return gPlatformFTLibrary;
 }
 #endif
+
+void
+gfxWindowsPlatform::InitDisplayCaps()
+{
+    HDC dc = GetDC((HWND)nsnull);
+
+    gfxPlatform::sDPI = GetDeviceCaps(dc, LOGPIXELSY);
+
+    ReleaseDC((HWND)nsnull, dc);
+}
+
