@@ -6,3 +6,8 @@ cp $1/include/sydney_audio.h include/sydney_audio.h
 cp $1/src/*.c src/
 cp $1/AUTHORS ./AUTHORS
 patch -p4 <pause-resume.patch
+patch -p4 <include-CoreServices.patch
+patch -p4 <sydney_os2_base.patch
+patch -p4 <sydney_os2_moz.patch
+patch -p3 <bug495794_closeAudio.patch
+patch -p3 < bug495558_alsa_endian.patch
