@@ -20,6 +20,7 @@
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s):
+ *  Josh Aas <josh@mozilla.com>
  *
  * Alternatively, the contents of this file may be used under the terms of
  * either the GNU General Public License Version 2 or later (the "GPL"), or
@@ -38,8 +39,6 @@
 #include "nsJSNPRuntime.h"
 #include "nsNPAPIPlugin.h"
 #include "nsNPAPIPluginInstance.h"
-#include "nsIPluginInstancePeer2.h"
-#include "nsPIPluginInstancePeer.h"
 #include "nsIScriptGlobalObject.h"
 #include "nsIScriptContext.h"
 #include "nsDOMJSUtils.h"
@@ -152,7 +151,7 @@ static JSBool
 NPObjWrapper_Construct(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
                        jsval *rval);
 
-static bool
+static JSBool
 CreateNPObjectMember(NPP npp, JSContext *cx, JSObject *obj,
                      NPObject *npobj, jsval id, jsval *vp);
 
@@ -295,11 +294,8 @@ GetJSContext(NPP npp)
   nsNPAPIPluginInstance *inst = (nsNPAPIPluginInstance *)npp->ndata;
   NS_ENSURE_TRUE(inst, nsnull);
 
-  nsCOMPtr<nsPIPluginInstancePeer> pp(do_QueryInterface(inst->Peer()));
-  NS_ENSURE_TRUE(pp, nsnull);
-
   nsCOMPtr<nsIPluginInstanceOwner> owner;
-  pp->GetOwner(getter_AddRefs(owner));
+  inst->GetOwner(getter_AddRefs(owner));
   NS_ENSURE_TRUE(owner, nsnull);
 
   nsCOMPtr<nsIDocument> doc;
@@ -354,7 +350,7 @@ NPVariantToJSVal(NPP npp, JSContext *cx, const NPVariant *variant)
   case NPVariantType_String :
     {
       const NPString *s = &NPVARIANT_TO_STRING(*variant);
-      NS_ConvertUTF8toUTF16 utf16String(s->utf8characters, s->utf8length);
+      NS_ConvertUTF8toUTF16 utf16String(s->UTF8Characters, s->UTF8Length);
 
       JSString *str =
         ::JS_NewUCStringCopyN(cx, reinterpret_cast<const jschar*>
@@ -1136,6 +1132,9 @@ GetNPObject(JSContext *cx, JSObject *obj)
   return (NPObject *)::JS_GetPrivate(cx, obj);
 }
 
+
+// Does not actually add a property because this is always followed by a
+// SetProperty call.
 static JSBool
 NPObjWrapper_AddProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
 {
@@ -1150,17 +1149,26 @@ NPObjWrapper_AddProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
 
   PluginDestructionGuard pdg(LookupNPP(npobj));
 
+  JSBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
+  if (!ReportExceptionIfPending(cx))
+    return JS_FALSE;
+
+  if (hasProperty)
+    return JS_TRUE;
+
   // We must permit methods here since JS_DefineUCFunction() will add
   // the function as a property
-  if (!npobj->_class->hasProperty(npobj, (NPIdentifier)id) &&
-      !npobj->_class->hasMethod(npobj, (NPIdentifier)id)) {
-    ThrowJSException(cx, "Trying to add unsupported property on scriptable "
-                     "plugin object!");
+  JSBool hasMethod = npobj->_class->hasMethod(npobj, (NPIdentifier)id);
+  if (!ReportExceptionIfPending(cx))
+    return JS_FALSE;
+
+  if (!hasMethod) {
+    ThrowJSException(cx, "Trying to add unsupported property on NPObject!");
 
     return JS_FALSE;
   }
 
-  return ReportExceptionIfPending(cx);
+  return JS_TRUE;
 }
 
 static JSBool
@@ -1177,7 +1185,11 @@ NPObjWrapper_DelProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
 
   PluginDestructionGuard pdg(LookupNPP(npobj));
 
-  if (!npobj->_class->hasProperty(npobj, (NPIdentifier)id))
+  JSBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
+  if (!ReportExceptionIfPending(cx))
+    return JS_FALSE;
+
+  if (!hasProperty)
     return JS_TRUE;
 
   if (!npobj->_class->removeProperty(npobj, (NPIdentifier)id))
@@ -1210,9 +1222,12 @@ NPObjWrapper_SetProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
 
   PluginDestructionGuard pdg(npp);
 
-  if (!npobj->_class->hasProperty(npobj, (NPIdentifier)id)) {
-    ThrowJSException(cx, "Trying to set unsupported property on scriptable "
-                     "plugin object!");
+  JSBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
+  if (!ReportExceptionIfPending(cx))
+    return JS_FALSE;
+
+  if (!hasProperty) {
+    ThrowJSException(cx, "Trying to set unsupported property on NPObject!");
 
     return JS_FALSE;
   }
@@ -1225,18 +1240,17 @@ NPObjWrapper_SetProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
   }
 
   JSBool ok = npobj->_class->setProperty(npobj, (NPIdentifier)id, &npv);
-
-  // Release the variant
-  _releasevariantvalue(&npv);
+  _releasevariantvalue(&npv); // Release the variant
+  if (!ReportExceptionIfPending(cx))
+    return JS_FALSE;
 
   if (!ok) {
-    ThrowJSException(cx, "Error setting property on scriptable plugin "
-                     "object!");
+    ThrowJSException(cx, "Error setting property on NPObject!");
 
     return JS_FALSE;
   }
 
-  return ReportExceptionIfPending(cx);
+  return JS_TRUE;
 }
 
 static JSBool
@@ -1263,9 +1277,14 @@ NPObjWrapper_GetProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
   PluginDestructionGuard pdg(npp);
 
   PRBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
-  PRBool hasMethod = npobj->_class->hasMethod(npobj, (NPIdentifier)id);
+  if (!ReportExceptionIfPending(cx))
+    return JS_FALSE;
 
-  // To support ambiguous members, we return NPObject Member class here.
+  PRBool hasMethod = npobj->_class->hasMethod(npobj, (NPIdentifier)id);
+  if (!ReportExceptionIfPending(cx))
+    return JS_FALSE;
+
+  // We return NPObject Member class here to support ambiguous members.
   if (hasProperty && hasMethod)
     return CreateNPObjectMember(npp, cx, obj, npobj, id, vp);
 
@@ -1273,22 +1292,16 @@ NPObjWrapper_GetProperty(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
     NPVariant npv;
     VOID_TO_NPVARIANT(npv);
 
-    if (!npobj->_class->getProperty(npobj, (NPIdentifier)id, &npv)) {
-      ThrowJSException(cx, "Error setting property on scriptable plugin "
-                       "object!");
+    if (npobj->_class->getProperty(npobj, (NPIdentifier)id, &npv))
+      *vp = NPVariantToJSVal(npp, cx, &npv);
 
-      return JS_FALSE;
-    }
-
-    *vp = NPVariantToJSVal(npp, cx, &npv);
-
-    // *vp now owns the value, release our reference.
     _releasevariantvalue(&npv);
 
-    return JS_TRUE;
+    if (!ReportExceptionIfPending(cx))
+      return JS_FALSE;
   }
 
-  return ReportExceptionIfPending(cx);
+  return JS_TRUE;
 }
 
 static JSBool
@@ -1412,7 +1425,10 @@ CallNPMethodInternal(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
   }
 
   if (!ok) {
-    ThrowJSException(cx, msg);
+    // ReportExceptionIfPending returns a return value, which is JS_TRUE
+    // if no exception was thrown. In that case, throw our own.
+    if (ReportExceptionIfPending(cx))
+      ThrowJSException(cx, msg);
 
     return JS_FALSE;
   }
@@ -1471,10 +1487,15 @@ NPObjWrapper_newEnumerate(JSContext *cx, JSObject *obj, JSIterateOp enum_op,
       enum_value = 0;
       length = 0;
     } else if (!npobj->_class->enumerate(npobj, &enum_value, &length)) {
-      ThrowJSException(cx, "Error enumerating properties on scriptable "
-                       "plugin object");
       delete state;
 
+      if (ReportExceptionIfPending(cx)) {
+        // ReportExceptionIfPending returns a return value, which is JS_TRUE
+        // if no exception was thrown. In that case, throw our own.
+        ThrowJSException(cx, "Error enumerating properties on scriptable "
+                             "plugin object");
+      }
+
       return JS_FALSE;
     }
 
@@ -1526,7 +1547,11 @@ NPObjWrapper_NewResolve(JSContext *cx, JSObject *obj, jsval id, uintN flags,
 
   PluginDestructionGuard pdg(LookupNPP(npobj));
 
-  if (npobj->_class->hasProperty(npobj, (NPIdentifier)id)) {
+  PRBool hasProperty = npobj->_class->hasProperty(npobj, (NPIdentifier)id);
+  if (!ReportExceptionIfPending(cx))
+    return JS_FALSE;
+
+  if (hasProperty) {
     JSBool ok;
 
     if (JSVAL_IS_STRING(id)) {
@@ -1545,7 +1570,15 @@ NPObjWrapper_NewResolve(JSContext *cx, JSObject *obj, jsval id, uintN flags,
     }
 
     *objp = obj;
-  } else if (npobj->_class->hasMethod(npobj, (NPIdentifier)id)) {
+
+    return JS_TRUE;
+  }
+
+  PRBool hasMethod = npobj->_class->hasMethod(npobj, (NPIdentifier)id);
+  if (!ReportExceptionIfPending(cx))
+    return JS_FALSE;
+
+  if (hasMethod) {
     JSString *str = nsnull;
 
     if (JSVAL_IS_STRING(id)) {
@@ -1572,7 +1605,8 @@ NPObjWrapper_NewResolve(JSContext *cx, JSObject *obj, jsval id, uintN flags,
     return fnc != nsnull;
   }
 
-  return ReportExceptionIfPending(cx);
+  // no property or method
+  return JS_TRUE;
 }
 
 static JSBool
@@ -1887,28 +1921,18 @@ nsJSNPRuntime::OnPluginDestroy(NPP npp)
   // Find the plugin instance so that we can (eventually) get to the
   // DOM element
   nsNPAPIPluginInstance *inst = (nsNPAPIPluginInstance *)npp->ndata;
-  if (!inst) {
+  if (!inst)
     return;
-  }
-
-  nsCOMPtr<nsIPluginInstancePeer> pip;
-  inst->GetPeer(getter_AddRefs(pip));
-  nsCOMPtr<nsIPluginTagInfo2> pti2(do_QueryInterface(pip));
-  if (!pti2) {
-    return;
-  }
 
   nsCOMPtr<nsIDOMElement> element;
-  pti2->GetDOMElement(getter_AddRefs(element));
-  if (!element) {
+  inst->GetDOMElement(getter_AddRefs(element));
+  if (!element)
     return;
-  }
 
   // Get the DOM element's JS object.
   nsCOMPtr<nsIXPConnect> xpc(do_GetService(nsIXPConnect::GetCID()));
-  if (!xpc) {
+  if (!xpc)
     return;
-  }
 
   // OK.  Now we have to get our hands on the right scope object, since
   // GetWrappedNativeOfNativeObject doesn't call PreCreate and hence won't get
@@ -1931,17 +1955,14 @@ nsJSNPRuntime::OnPluginDestroy(NPP npp)
     return;
   }
 
-  nsCOMPtr<nsISupports> supp(do_QueryInterface(element));
-  nsCOMPtr<nsIXPConnectWrappedNative> holder;
-  xpc->GetWrappedNativeOfNativeObject(cx, sgo->GetGlobalJSObject(), supp,
-                                      NS_GET_IID(nsISupports),
-                                      getter_AddRefs(holder));
-  if (!holder) {
+  nsCOMPtr<nsINode> node(do_QueryInterface(element));
+
+  JSObject *obj;
+  if (!node || !(obj = node->GetWrapper())) {
     return;
   }
 
-  JSObject *obj, *proto;
-  holder->GetJSObject(&obj);
+  JSObject *proto;
 
   // Loop over the DOM element's JS object prototype chain and remove
   // all JS objects of the class sNPObjectJSWrapperClass (there should
@@ -1985,23 +2006,23 @@ LookupNPP(NPObject *npobj)
   return entry->mNpp;
 }
 
-bool
+JSBool
 CreateNPObjectMember(NPP npp, JSContext *cx, JSObject *obj,
                      NPObject* npobj, jsval id, jsval *vp)
 {
-  NS_ENSURE_TRUE(vp, false);
+  NS_ENSURE_TRUE(vp, JS_FALSE);
 
   if (!npobj || !npobj->_class || !npobj->_class->getProperty ||
       !npobj->_class->invoke) {
     ThrowJSException(cx, "Bad NPObject");
 
-    return false;
+    return JS_FALSE;
   }
 
   NPObjectMemberPrivate *memberPrivate =
     (NPObjectMemberPrivate *)PR_Malloc(sizeof(NPObjectMemberPrivate));
   if (!memberPrivate)
-    return false;
+    return JS_FALSE;
 
   // Make sure to clear all members in case something fails here
   // during initialization.
@@ -2010,7 +2031,7 @@ CreateNPObjectMember(NPP npp, JSContext *cx, JSObject *obj,
   JSObject *memobj = ::JS_NewObject(cx, &sNPObjectMemberClass, nsnull, nsnull);
   if (!memobj) {
     PR_Free(memberPrivate);
-    return false;
+    return JS_FALSE;
   }
 
   *vp = OBJECT_TO_JSVAL(memobj);
@@ -2021,9 +2042,17 @@ CreateNPObjectMember(NPP npp, JSContext *cx, JSObject *obj,
   jsval fieldValue;
   NPVariant npv;
   VOID_TO_NPVARIANT(npv);
-  if (!npobj->_class->getProperty(npobj, (NPIdentifier)id, &npv)) {
+
+  NPBool hasProperty = npobj->_class->getProperty(npobj, (NPIdentifier)id,
+                                                  &npv);
+  if (ReportExceptionIfPending(cx)) {
     ::JS_RemoveRoot(cx, vp);
-    return false;
+    return JS_FALSE;
+  }
+
+  if (!hasProperty) {
+    ::JS_RemoveRoot(cx, vp);
+    return JS_FALSE;
   }
 
   fieldValue = NPVariantToJSVal(npp, cx, &npv);
@@ -2043,7 +2072,7 @@ CreateNPObjectMember(NPP npp, JSContext *cx, JSObject *obj,
 
   ::JS_RemoveRoot(cx, vp);
 
-  return true;
+  return JS_TRUE;
 }
 
 static JSBool
@@ -2053,7 +2082,10 @@ NPObjectMember_Convert(JSContext *cx, JSObject *obj, JSType type, jsval *vp)
     (NPObjectMemberPrivate *)::JS_GetInstancePrivate(cx, obj,
                                                      &sNPObjectMemberClass,
                                                      nsnull);
-  NS_ASSERTION(memberPrivate, "no Ambiguous Member Private data!");
+  if (!memberPrivate) {
+    NS_ERROR("no Ambiguous Member Private data!");
+    return JS_FALSE;
+  }
 
   switch (type) {
   case JSTYPE_VOID:
@@ -2149,7 +2181,10 @@ NPObjectMember_Call(JSContext *cx, JSObject *obj,
   }
 
   if (!ok) {
-    ThrowJSException(cx, "Error calling method on NPObject!");
+    // ReportExceptionIfPending returns a return value, which is JS_TRUE
+    // if no exception was thrown. In that case, throw our own.
+    if (ReportExceptionIfPending(cx))
+      ThrowJSException(cx, "Error calling method on NPObject!");
 
     return JS_FALSE;
   }
